////////////////////////////////////////////////////////////////////////////////////////////////////
//
// MIDI2LED 
//
// A short program that receives MIDI NoteOn / NoteOff signals on the RX pin, and controls a
// W2812B LED strip according to the keys pressed. The color of the LEDs changes automatically
// through time.
// Written 2020 by frankje
//
// - An entry to the Intructibles LED strip challenge 2020 -
//
////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////
// 
// CHANGE THESE ACCORDING TO YOUR SETUP
//
///////////////////////////////////////////////


// The number of LEDs in your W2812B LED strip
#define NUMBER_OF_LEDS 60

// The number of keys on your keyboard
#define NUMBER_OF_KEYS 76

// The MIDI note number of the lowest key on your keyboard
#define MIN_KEY 28

///////////////////////////////////////////////
//
///////////////////////////////////////////////


#include <MIDI.h>

MIDI_CREATE_DEFAULT_INSTANCE();

#include <Adafruit_NeoPixel.h>
//#include <avr/power.h>

// The I/O pin that the data line of the LED strip is connected to
#define LED_STRIP_PIN 6

// Arduino Board LED is on Pin 13
#define LED_PIN 13    

Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMBER_OF_LEDS, LED_STRIP_PIN, NEO_GRB + NEO_KHZ800);

// The maximum value the deltaRgb elements can assume. The higher the value, the faster the color
// can change 
#define MAX_COLOR_CHANGE_SPEED 20

// an array of last received MIDI velocities for the different keys of
// the keyboard. A velocity of zero means a MIDI off command has been sent
byte velocityArray[NUMBER_OF_KEYS];

// the velocity array, mapped to the LEDs. Each entry is the maximum of the velocities
// of the keys mapped to that LED.
byte ledArray[NUMBER_OF_LEDS];

// which LEDs need to be updated this time? Saves time when multiplying velocities with basic colors
bool ledNeedsUpdate[NUMBER_OF_LEDS];

struct rgb_struct
{
  word red;
  word green;
  word blue;
} ;

// stores the basic colors for all LEDs [0...32*256]^3 (ignore the lower 6 bits for each channel). 
// Has to be multiplied with a velocity and then scaled down to get the actual RGB value
rgb_struct colorArray[NUMBER_OF_LEDS];

// the basic color of each LED, scaled with the velocity. This goes directly into the pixel structure
rgb_struct ledColors[NUMBER_OF_LEDS];

// a mapping from keys to LEDs
byte ledForKey[NUMBER_OF_KEYS];

// the r/g/b direction in which the current colour is moving
int deltaRgb[3] = {-1,-1,-1}; 

// the current colour (ignore the lower 6 bits)
word currentRgb[3]= {0,0,0}; 

// used for decayVelocities
unsigned long timeOfLastDecay = 0;



void initLedForKey()
{
  for (int keyNumber=0; keyNumber<NUMBER_OF_KEYS; keyNumber++)
  {
    ledForKey[keyNumber] = (byte) map(keyNumber, 
      0, NUMBER_OF_KEYS-1, 
      0, NUMBER_OF_LEDS-1);
  }

}

///////////////////////////////////
//                               //
// VELOCITY ARRAY                //
//                               //
///////////////////////////////////

void initVelocityArray()
{
  for (int i = 0; i<NUMBER_OF_KEYS; i++)
  {
    velocityArray[i] = 0;
  }
}


void updateVelocityArray(int pitch, byte velocity)
// stores a new velocity for a key. Also updates the velocity for the LED mapped to that key, and the color,
// if the new velocity is higher
{
  byte keyNumber = pitch - MIN_KEY;
  if ((keyNumber >= 0) && (keyNumber < NUMBER_OF_KEYS))
  {
    int ledNumber = ledForKey[keyNumber]; 

    // if this note is louder than what is already in the velocity array ...
    if (velocity >= ledArray[ledNumber])
    {
      // ... then update the LED with the current color
      colorArray[ledNumber].red = currentRgb[0];
      colorArray[ledNumber].green = currentRgb[1];
      colorArray[ledNumber].blue = currentRgb[2];
    }
    //  and update the velocity arrays
    velocityArray[keyNumber] = velocity;
    ledArray[ledNumber] = velocity;
    
    // mark this LED for update
    ledNeedsUpdate[ledNumber] = true;
  
  }
}


///////////////////////////////////
//                               //
// "CURRENT COLOR" GENERATOR     //
//                               //
///////////////////////////////////

void updateCurrentRgb()
// change the current Rgb by deltaRgb; in case an upper or lower limit is reached, get new deltaRgb
{
  bool changeDirection = false;
  bool positiveDirectionAllowed[3] = {true, true, true};
  bool negativeDirectionAllowed[3] = {true, true, true};
   
  for (int i=0; i<3; i++)
  {
    currentRgb[i] = currentRgb[i] + deltaRgb[i];
    if (currentRgb[i] <= 0)
    {
      negativeDirectionAllowed[i] = false;
      changeDirection = true;
      currentRgb[i] = 0;
    }
    
    if (currentRgb[i] >= 255*32)
    {
      positiveDirectionAllowed[i] = false;
      changeDirection = true;
      currentRgb[i] = 255*32;
    }
    
  }

  if (changeDirection)
  {
    for (int i=0; i<3; i++)
    {
      int newDir = random(1, MAX_COLOR_CHANGE_SPEED); // the 1 ensures that the new direction can never be zero, i.e. the system cannot get stuck
      int negative = random(2); // 0 = positive, 1= negative

      // don't move into a direction that is not allowed
      if (not(negativeDirectionAllowed[i]))
      {
        negative = 0;
      }
      if (not(positiveDirectionAllowed[i]))
      {
        negative = 1;
      }

      if (negative == 1)
      {
        newDir = -newDir;
      }

      deltaRgb[i] = newDir;
    }
    
  }


}



void decayVelocities()
// decrease all velocities in velocityArray and ledArray by 1,
// if greater than zero
{
  for (int keyNumber= 0; keyNumber<NUMBER_OF_KEYS; keyNumber++)
  {
    if (velocityArray[keyNumber] > 0)
    {
      velocityArray[keyNumber]--;
    }
  }
  
  for (int ledNumber = 0; ledNumber<NUMBER_OF_LEDS; ledNumber++)
  {
    if (ledArray[ledNumber] > 0)
    {
      ledArray[ledNumber]--;
      // mark this LED for update
      ledNeedsUpdate[ledNumber] = true;
    }
    
  }
}




///////////////////////////////////
//                               //
// LED COLOR UPDATING            //
//                               //
///////////////////////////////////

void updateLedArray()
// 
{
  // change the current color
  updateCurrentRgb();

  // Every 20 milliseconds, decay velocities
  if (millis() > timeOfLastDecay + 20)
  {
    decayVelocities();
    timeOfLastDecay = millis();
  }

  // for all LEDs marked for update, compute the new color from the basic color and the velocity
  for (int ledNumber = 0; ledNumber < NUMBER_OF_LEDS; ledNumber++)
  {
    rgb_struct rgb;
    word velocity6bit;
    if (ledNeedsUpdate[ledNumber])
    {
      rgb = colorArray[ledNumber];

      // This maps the [0..32*255] color to [0..8*255]
      rgb.red= rgb.red/4;
      rgb.green= rgb.green/4;
      rgb.blue= rgb.blue/4;
      
      // this maps the [0..127] velocity to [0..31]
      velocity6bit = ledArray[ledNumber]/4; 
      
      ledColors[ledNumber].red = (rgb.red * velocity6bit)/256;
      ledColors[ledNumber].green = (rgb.green * velocity6bit)/256;
      ledColors[ledNumber].blue = (rgb.blue * velocity6bit)/256;

      ledNeedsUpdate[ledNumber] = false;
    }
  }
  
}   

void showLedArray()
{
  for(int i=0;i<NUMBER_OF_LEDS;i++){
    pixels.setPixelColor(i, pixels.Color(ledColors[i].red,ledColors[i].green,ledColors[i].blue)); 
  }
  
  pixels.show();
}

void initLedArray(byte val)
{
  for (int i=0; i<NUMBER_OF_LEDS; i++)
  {
    ledArray[i] = val;
  }
}



///////////////////////////////////
//                               //
// MIDI HANDLES                  //
//                               //
///////////////////////////////////

void MyHandleNoteOn(byte channel, byte pitch, byte velocity) 
// This is called by the MIDI library when a MIDI NoteOn message is received.
// It flasheds the onboard LED and sends the velocity to the velocityArray
{
  digitalWrite(LED_PIN, HIGH);  //Turn LED on
  if (velocity == 0) 
  {
    // A NoteOn message with a velocity = Zero is actualy a NoteOff
    digitalWrite(LED_PIN, LOW);//Turn LED off
  }
  updateVelocityArray(pitch, velocity);
  
}

void MyHandleNoteOff(byte channel, byte pitch, byte velocity)
// This is called by the MIDI library when a MIDI NoteOff message is received.
// It turns off the onboard LED and sends the velocity to the velocityArray
{
  digitalWrite(LED_PIN, LOW);  //Turn LED off
  
  updateVelocityArray(pitch, 0);
}

///////////////////////////////////
//                               //
// SETUP & MAIN LOOP             //
//                               //
///////////////////////////////////

void setup() {

  pinMode (LED_PIN, OUTPUT); // Set Arduino board pin 13 to output - will flash on MIDI NoteOn event

  // setup the structure used to address the LED strip
  pixels.begin();

  // Test all LEDs on your strip
  for (int i=0; i<NUMBER_OF_LEDS; i++)
  {
    pixels.setPixelColor(i, pixels.Color(255,0,0));
    pixels.show();
    delay(10);
    pixels.setPixelColor(i, pixels.Color(0,0,0));
    
  }
  for (int i=0; i<NUMBER_OF_LEDS; i++)
  {
    pixels.setPixelColor(i, pixels.Color(0,255,0));
    pixels.show();
    delay(10);
    pixels.setPixelColor(i, pixels.Color(0,0,0));
    
  }
  for (int i=0; i<NUMBER_OF_LEDS; i++)
  {
    pixels.setPixelColor(i, pixels.Color(0,0,255));
    pixels.show();
    delay(10);
    pixels.setPixelColor(i, pixels.Color(0,0,0));
    
  }
  pixels.show();

  // initialize the mapping from MIDI note number to LED number
  initLedForKey();

  // Initialize the Midi Library.
  // OMNI sets it to listen to all channels.
  MIDI.begin(MIDI_CHANNEL_OMNI); 
  
  // Assign the function to be called when a MIDI NoteOn command is received
  MIDI.setHandleNoteOn(MyHandleNoteOn); 

  // Assign the function to be called when a MIDI NoteOff command is received
  MIDI.setHandleNoteOff(MyHandleNoteOff);

  // Set all LED velocities to zero, i.e. all LEDs are off
  initLedArray(0);

}

void loop() 
{
  MIDI.read(); // Continuously check what Midi Commands have been received.

  updateLedArray();
  
  showLedArray();
}
